/*
 * hfcl_diag.c
 * Copyright (C) 2007, 2010, Hitachi, Ltd.
 * Author(s): Yoshihiro Toyohara <yoshihiro.toyohara.qs@hitachi.com>
 */

char diag_rcsid[] = "$Id: hfcl_diag.c,v 1.32.2.10.2.3.4.4 2010/05/22 01:56:17 eteshima Exp $";

#include "hfcldd.h"
#include "hfcl_detect.h"
#include "hfcl_top.h"
#include "hfcl_ioctl.h"
#include "hfcl_timer_recovery.h"
#include "hfcl_hand_timer_trace.h"
#include "hfcl_mlpf.h"
#include "hfcl_version.h"

/* Log size */
#define LOG_LENGTH_LDCH			0x8000	/* Load CH Trace Log(32Kbyte) */
#define LOG_LENGTH_MEINT		0x600	/* Send MEINT Log(1.5Kbyte)  */
#define LOG_LENGTH_SOFT			0x12000	/* Forced Soft Log(72Kbyte)  */

/* Log area segment number */
#define LOG_SEGNUM_LDCH			8	/* Load CH Trace Log */
#define LOG_SEGNUM_MEINT		1	/* Send MEINT Log  */
#define LOG_SEGNUM_SOFT			18	/* Forced Soft Log */

#define FPP_TRC_SIZE			32	/* Load CH Trace LOG(FPP Trace Pointer size) */
#define FRM_TRC_SIZE			8	/* Load CH Trace LOG(Frame Trace Pointer size) */

#define FPP_TRC_SIZE_OFFSET		8	/* Load CH Trace LOG(Response head offset of FPP Trace Pointer) */
#define FRM_TRC_SIZE_OFFSET		40	/* Load CH Trace LOG(Response head offset of Frame Trace Pointer) */

#define MAX_READ_CNT			1	/* Maximum Read Count */


/* PCI access */
#define PCI_LENGTH_01			0x01	/* Access size  1byte */
#define PCI_LENGTH_02			0x02	/* Access size  2byte */
#define PCI_LENGTH_04			0x04	/* Access size  4byte */
#define PCI_ADDR_FPP_MAX		0xfff	/* FPP PCI memory space address(Maximum) */
#define PCI_ADDR_FIVE_MAX		0x1fff	/* FIVE PCI memory space address(Maximum) */
#define PCI_CNF_ADDR_MAX		0xff	/* PCI configuration space address(Maximum) */


/*-- global variable --*/
extern int mih_log(struct adap_info *ap,struct diag_ioctl *diag);
extern int load_ch_trace_log(struct adap_info *ap,struct diag_ioctl *diag);
extern int forced_log(struct adap_info *ap,struct diag_ioctl *diag);
extern int fw_start(struct adap_info *ap,struct diag_ioctl *diag);
extern int fw_post(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_fw_init_tbl(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_xob(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_xrb(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_seg_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_adap_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_target_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_hfctrace(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_mailbox(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_version(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_hwlog(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_tree(struct adap_info *ap,struct diag_ioctl *diag);
extern int pci_access(struct adap_info *ap,struct diag_ioctl *diag);
extern int pci_cnf_access(struct adap_info *ap,struct diag_ioctl *diag);
extern int stop_func(struct adap_info *ap,struct diag_ioctl *diag);
extern int init_mode_set(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_mpadap_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int fcp_mode_set(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_manage_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_hfcpkt(struct adap_info *ap,struct diag_ioctl *diag);	/*2004.11.30*/
extern int read_dev_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_lg_target_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_lg_dev_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_lg_path_info(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_lg_path_info1(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_lg_path_info2(struct adap_info *ap,struct diag_ioctl *diag);
extern int read_failover_info(struct adap_info *ap,struct diag_ioctl *diag);

const Typ_diag_cntl diag_check[]={
/*----- Sub_Command ------------- Mode --------------- FPP-Flag ---- FIVE-flag ---*/
	{ HFC_DG_MIHLOG   , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xc0008000 , 0xc0008000 , (void*)mih_log },
	{ HFC_DG_LDCHTRC  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xc0008000 , 0xc0008000 , (void*)load_ch_trace_log },
	{ HFC_DG_FORCSLOG , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xc0000000 , 0xf0000000 , (void*)forced_log },
	{ HFC_DG_FWSTART  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)fw_start },
	{ HFC_DG_FWPOST   , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0x40008000 , 0x40008000 , (void*)fw_post },
	{ HFC_DGRD_INITTBL, HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_fw_init_tbl},
	{ HFC_DGRD_XOB	  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_xob },
	{ HFC_DGRD_XRB	  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_xrb },
	{ HFC_DGRD_SEGINFO, HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_seg_info },
	{ HFC_DGRD_ADAP   , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_adap_info },
	{ HFC_DGRD_TARGET , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_target_info },
	{ HFC_DGRD_DDTRC  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0008000 , 0xf0008000 , (void*)read_hfctrace },
	{ HFC_DGRD_MB     , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_mailbox },
	{ HFC_DGRD_VER    , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_version },
	{ HFC_DGRD_HWLOG  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_hwlog },
	{ HFC_DGRD_TREE   , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_tree },
	{ HFC_DGRD_SCMD   , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_hfcpkt }, /*2004.11.30 */
	{ HFC_DG_RDPCI    , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)pci_access },
	{ HFC_DG_WRPCI    , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)pci_access },
	{ HFC_DG_RDPCI_CFG, HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)pci_cnf_access },
	{ HFC_DG_WRPCI_CFG, HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)pci_cnf_access },
	{ HFC_DG_FSTOP    , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)stop_func },
	{ HFC_DGRD_MPINFO , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_mpadap_info },
	{ HFC_DG_INITMDST , HFC_DIAG_FIVE                , 0          , 0xb0000000 , (void*)init_mode_set },
	{ HFC_DG_FCPMDST  , HFC_DIAG_FIVE                , 0          , 0x30000000 , (void*)fcp_mode_set },
	{ HFC_DGRD_MNGINFO, HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_manage_info },
	{ HFC_DGRD_DEV    , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_dev_info },
	{ HFC_DGRD_LGTGT  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_lg_target_info },
	{ HFC_DGRD_LGDEV  , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_lg_dev_info },
	{ HFC_DGRD_PATHINFO, HFC_DIAG_FPP| HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_lg_path_info },
	{ HFC_DGRD_PATHINFO1,HFC_DIAG_FPP| HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_lg_path_info1 },
	{ HFC_DGRD_PATHINFO2,HFC_DIAG_FPP| HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_lg_path_info2 },
	{ HFC_DGRD_FOINFO , HFC_DIAG_FPP | HFC_DIAG_FIVE , 0xf0000000 , 0xf0000000 , (void*)read_failover_info },
	{ 0               , 0                            , 0          , 0          , NULL }
};
/*--------------------------------------------------------------*
 * Log address list
 * Notes:
 *       Load ch Trace Log : log_addr[0-7] 
 *       Send MEINT Log    : log_addr[0]
 *       Forced soft Log   : log_addr[0-17] 
 *--------------------------------------------------------------*/
struct logaddr_list {
	uchar	segnum;				/* Number of segments */
	uchar	resv0;
	uchar	resv1;
	uchar	resv2;
	uint	resv3;
	uint64_t	bus_addr[18];	/* Bus address */
};

/*
 * Function:    set_fw_trace_mode
 *
 * Purpose:     Process for ioctl (HFCDIAG0) subcommand (Set FW Trace Mode) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int set_fw_trace_mode(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{

	int		loop;
	int		rtn;

	HFC_ENTRY("set_fw_trace_mode") ;
	/*--------------------------------------------------------------*
	 * Mailbox processing
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock mailbox */

	/*--------------------------------------------------------------*
	 * Set specified trace information to fw_init_tbl
	 *--------------------------------------------------------------*/
	BCOPY( (char *)&diag->uni.trc_mode , (char *)&ap->fw_init_p->trc_info, sizeof(struct fw_trc_info ));
	for( loop = 0; loop < 4; loop++ ){
		hfc_write_val( ap->fw_init_p->trc_info.trc_seg[loop].top_port_id, diag->uni.trc_mode.trc_seg[loop].top_port_id );
		hfc_write_val( ap->fw_init_p->trc_info.trc_seg[loop].bottom_port_id, diag->uni.trc_mode.trc_seg[loop].bottom_port_id );
	}

	/*--------------------------------------------------------------*
	 * Set up and initiate mailbox
	 *--------------------------------------------------------------*/
	/* Set up mailbox control block  */
	ap->mb->mb_init.command = HFC_MBCMD_LOGTRACE;	/* Command */
	ap->mb->mb_init.sub_cmd = HFC_MBSCMD_SETFWTR;	/* Subcommand */
	hfc_write_val( ap->mb->mb_init.dependent_code, 0x0000 );

	/* Initiate mailbox and wait for completion */
	rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry);	/*	FCLNX-0523 */
	unlock_mailbox( ap );	/* Unlock mailbox */

	HFC_EXIT("set_fw_trace_mode") ;
	return( rtn );
}


/*
 * Function:    mih_log
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (MIH Log)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int mih_log(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int		rtn;

	uchar	ssn = 0 ;					/* slog page No 	*/
	uchar	son = 0 ;					/* slog offset No	*/
	ushort	sbc = 0 ;					/* slog length		*/
	ushort	usSlogLen;
	ushort	usSlogNum;

	uint	slog_in_page ;
	uint	slog_adr ;

	int copy_cnt = 0 ;
	int i ;

	HFC_ENTRY("mih_log") ;

	/*--------------------------------------------------------------*
	 * Mailbox processing
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock mailbox */

	/* Set up mailbox control block  */
	ap->mb->mb_init.command = HFC_MBCMD_LOGTRACE;	/* Command */
	ap->mb->mb_init.sub_cmd = HFC_MBSCMD_MIHLOG;	/* Sub-Command */
	hfc_write_val( ap->mb->mb_init.dependent_code, 0x0010 );	/* LOGID */

	/* Mailbox start/completion waiting */
	if (( rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry) ) == 0 ) {	/* FCLNX-0523 */
		/* Copy data to Soft_Log_Area designated by ssn/son/sbc */
		ssn = (uchar)hfc_read_val( ap->mb->mb_resp.type.drvlogb0.mih_log.ssn );
		son = (uchar)hfc_read_val( ap->mb->mb_resp.type.drvlogb0.mih_log.son );
		sbc = (ushort)hfc_read_val( ap->mb->mb_resp.type.drvlogb0.mih_log.sbc );
		usSlogLen = (ushort)hfc_read_val( ap->fw_init_p->slog_len );
		usSlogNum = (ushort)hfc_read_val( ap->fw_init_p->slog_num );

		if( sbc != 0 )
		{
			slog_in_page = HFC_PAGE_SIZE / (uint)usSlogLen ;
			if( (son < slog_in_page) && (ssn < usSlogNum) )
			{
				if ( sbc > usSlogLen )	/* sbc>1024 */
				{
					/* If the sbc is larger than 1024, copy every 1024 bytes */
					copy_cnt = sbc / usSlogLen;	/* Sbc/1024 */
					if( sbc % usSlogLen )		/* When residue exists, count up copy_cnt */
						copy_cnt++;

					for( i = 0 ; i < copy_cnt ; i++ )
					{
						slog_adr = (HFC_PAGE_SIZE * ssn)+(usSlogLen * son) ;

						if( i == (copy_cnt-1) )
							/* Copy sbc-(1024*i) bytes data from ap->slog[slog_adr] to diag->addr+(1024*i) */
							COPYOUT( &ap->slog[slog_adr], ( (uchar *)(ulong)diag->addr+(usSlogLen*i) ), ( sbc - (usSlogLen*i) ) );
						else
							/* Copy 1024 bytes data from ap->slog[slog_adr] to diag->addr+(1024*i) */
							COPYOUT( &ap->slog[slog_adr], ( (uchar *)(ulong)diag->addr+(usSlogLen*i) ), usSlogLen );

						if( ( ssn == (usSlogNum-1) ) &&
							( son == (slog_in_page-1) ) )	/* ssn=63, son=3 */
						{
							ssn = 0;
							son = 0;
						}
						else
						{
							if( son == (slog_in_page-1) )	/* son=3 */
							{
								ssn++;
								son = 0;
							}
							else
								son++;
						}
					}
				}
				else
				{
					/* If sbc is 1024 or less, copy it at a time */
					slog_adr = (HFC_PAGE_SIZE * ssn)+(usSlogLen * son) ;
					COPYOUT( &ap->slog[slog_adr], (uchar *)(ulong)diag->addr, sbc );
				}
			}
		}

		/* Normal end without outputting data to buffer when ssn/son/sbc check is error  */
	}

	unlock_mailbox( ap );	/* Lock release of mailbox */


	HFC_EXIT("mih_log") ;

	return( rtn );
}


/*
 * Function:    load_ch_trace_log
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Load CH Trace Log)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int load_ch_trace_log(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	struct	logaddr_list *loglist;			/* Log address list virtual address */
	dma_addr_t	logaddr_busaddr;			/* Log address list bus address */
	char	*trc_log_area;					/* Log buffer area virtual address */
	dma_addr_t	trclog_busaddr;				/* Log buffer area bus address */

	char	*work_adr;						/* work pointer address */
	ushort	uswork;							/* work */

	int		i,j,k,rtn;
	int		loop_cnt;

	uchar	errlog_no ;

	HFC_ENTRY("load_ch_trace_log") ;
	/*--------------------------------------------------------------*
	 * Log area specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unallocated ? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_LDCHTRC, 0x00 );
		return( EINVAL );
	}
	if( diag->length != LOG_LENGTH_LDCH ) {	/* Is length 32K bytes ? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_LDCHTRC, 0x01 );
		return( EINVAL );
	}
	/*--------------------------------------------------------------*
	 * Allocate logout list area 
	 *--------------------------------------------------------------*/

	loglist = (struct logaddr_list *)pci_alloc_consistent(ap->pci_cfginf,
		(uint)sizeof(struct logaddr_list), &logaddr_busaddr);

	if (loglist == NULL) {
		errlog_no = 0x05 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x60, &errlog_no, 1) ;
		return ( ENOMEM );
	}

	BZERO((char *)loglist, (uint)sizeof(struct logaddr_list));

	/*--------------------------------------------------------------*
	 * Allocate log buffer area
	 *--------------------------------------------------------------*/
	/* Internal partitioning & page mapping */
	trc_log_area = (uchar *)pci_alloc_consistent(ap->pci_cfginf,
		LOG_LENGTH_LDCH, &trclog_busaddr);

	if (trc_log_area == NULL) {
		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);

		errlog_no = 0x06 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x60, &errlog_no, 1) ;
		return ( ENOMEM );
	}

	BZERO((char *)trc_log_area, LOG_LENGTH_LDCH);

	/*--------------------------------------------------------------*
	 * Set up log addresses to log list 
	 *--------------------------------------------------------------*/
	loglist->segnum = LOG_SEGNUM_LDCH;		/* Total number of segments */

	k = 0 ;
	loop_cnt = 0 ;

	if ( LOG_LENGTH_LDCH % HFC_PAGE_SIZE ) {
		/* If residue exists, loop number is (Page number + 1) */
		loop_cnt = (LOG_LENGTH_LDCH / HFC_PAGE_SIZE) + 1 ;
	}
	else {
		/* If no residue exists, loop number is page number */
		loop_cnt = LOG_LENGTH_LDCH / HFC_PAGE_SIZE ;
	}

	/* Set bus address to log list */
	for ( j = 0 ; j < loop_cnt ; j++) {
		hfc_write_val(loglist->bus_addr[k], (uint64_t)(trclog_busaddr + HFC_PAGE_SIZE*j) );

		if ( k >= LOG_SEGNUM_LDCH ) {	
			pci_free_consistent(ap->pci_cfginf,
				(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
			pci_free_consistent(ap->pci_cfginf,
				LOG_LENGTH_LDCH, (void *)trc_log_area, trclog_busaddr);

			return ( EIO );
		}

		k++ ;
	}

	/*--------------------------------------------------------------*
	 * Mailbox processing
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock mailbox */

	/* Set up mailbox control block  */
	ap->mb-> mb_init.command = HFC_MBCMD_LOGTRACE;	/* Command */
	ap->mb-> mb_init.sub_cmd = HFC_MBSCMD_LDCHTRC;	/* Sub-Command */
	hfc_write_val( ap->mb->mb_init.dependent_code, 0x0000 );
	hfc_write_val( ap->mb->mb_init.type.drvlogb1.log_list_addr,
			(uint64_t)logaddr_busaddr );/* Bus address of log address list */

	/* Mailbox start/completion waiting */
	if ((rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry)) != 0) {	/* FCLNX-0523 */
		/* Error */
		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
		pci_free_consistent(ap->pci_cfginf,
			LOG_LENGTH_LDCH, (void *)trc_log_area, trclog_busaddr);

		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_LDCHTRC, 0x00 );

		unlock_mailbox( ap );	/* Unlock mailbox */
		return ( rtn );
	}

	/* Set trace pointer */
	work_adr = (char *)&ap->mb->mb_resp.type.drvlogb0.mih_log;
	BCOPY( (char *)(work_adr + FPP_TRC_SIZE_OFFSET),
		(char *)&diag->uni.ch_trc_log.trc_ptr[0], FPP_TRC_SIZE );
	for( i = 0; i < 4; i++ ){
		uswork = (ushort)hfc_read_val( diag->uni.ch_trc_log.trc_ptr[i].sp );
		diag->uni.ch_trc_log.trc_ptr[i].sp = uswork;
		uswork = (ushort)hfc_read_val( diag->uni.ch_trc_log.trc_ptr[i].ep );
		diag->uni.ch_trc_log.trc_ptr[i].ep = uswork;
		uswork = (ushort)hfc_read_val( diag->uni.ch_trc_log.trc_ptr[i].cp );
		diag->uni.ch_trc_log.trc_ptr[i].cp = uswork;
	}

	BCOPY( (char *)(work_adr + FRM_TRC_SIZE_OFFSET),
		(char *)&diag->uni.ch_trc_log.frm_trc_ptr, FRM_TRC_SIZE );
	uswork = (ushort)hfc_read_val( diag->uni.ch_trc_log.frm_trc_ptr.sp );
	diag->uni.ch_trc_log.frm_trc_ptr.sp = uswork;
	uswork = (ushort)hfc_read_val( diag->uni.ch_trc_log.frm_trc_ptr.ep );
	diag->uni.ch_trc_log.frm_trc_ptr.ep = uswork;
	uswork = (ushort)hfc_read_val( diag->uni.ch_trc_log.frm_trc_ptr.cp );
	diag->uni.ch_trc_log.frm_trc_ptr.cp = uswork;

	unlock_mailbox( ap );	/* Unlock mailbox */

	/*--------------------------------------------------------------*
	 * Data copy 
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)trc_log_area, (char *)(ulong)diag->addr, LOG_LENGTH_LDCH ) != 0 ) {

		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
		pci_free_consistent(ap->pci_cfginf,
			LOG_LENGTH_LDCH, (void *)trc_log_area, trclog_busaddr);

		errlog_no = 0x0d ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	pci_free_consistent(ap->pci_cfginf,
		(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
	pci_free_consistent(ap->pci_cfginf,
		LOG_LENGTH_LDCH, (void *)trc_log_area, trclog_busaddr);

	HFC_EXIT("load_ch_trace_log");
	return( rtn );

}


/*
 * Function:    meint_log
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Send MEINT Log) 
 *
 * Arguments:   
 *  ap         - struct adap_info 
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int meint_log(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	struct	logaddr_list *loglist;			/* Log address list virtual address */
	dma_addr_t	logaddr_busaddr;			/* Log address list bus address */
	char	*trc_log_area;					/* Log buffer area virtual address */
	dma_addr_t	logout_busaddr;				/* Log buffer area bus address */

	int		rtn;

	uchar	errlog_no ;

	HFC_ENTRY("meint_log") ;

	/*--------------------------------------------------------------*
	 * Log area specification check (one segment?)
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	 /* Is virtual address unallocated ? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MEINTLOG, 0x00 );
		return( EINVAL );
	}
	if( diag->length != LOG_LENGTH_MEINT ) {/* Is length 1.5K bytes ? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MEINTLOG, 0x01 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Allocate logout list area
	 *--------------------------------------------------------------*/
	
	loglist = (struct logaddr_list *)pci_alloc_consistent(ap->pci_cfginf,
		(uint)sizeof(struct logaddr_list), &logaddr_busaddr);

	if (loglist == NULL) {
		errlog_no = 0x0e ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x60, &errlog_no, 1) ;
		return ( ENOMEM );
	}

	BZERO((char *)loglist, (uint)sizeof(struct logaddr_list));

	/*--------------------------------------------------------------*
	 * Allocate log buffer area
	 *--------------------------------------------------------------*/

	trc_log_area = (uchar *)pci_alloc_consistent(ap->pci_cfginf,
		LOG_LENGTH_MEINT, &logout_busaddr);

	if (trc_log_area == NULL) {
		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);

		errlog_no = 0x11 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x60, &errlog_no, 1) ;
		return ( ENOMEM );
	}

	BZERO((char *)trc_log_area, LOG_LENGTH_MEINT);

	/*--------------------------------------------------------------*
	 * Set up log addresses to log list
	 *--------------------------------------------------------------*/
	loglist->segnum = LOG_SEGNUM_MEINT;		/* Number of segments */
	hfc_write_val(loglist->bus_addr[0], (uint64_t)logout_busaddr );

	/*--------------------------------------------------------------*
	 * Mailbox processing
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock mailbox */

	/* Set up mailbox control block  */
	ap->mb->mb_init.command = HFC_MBCMD_LOGTRACE;	/* Command */
	ap->mb->mb_init.sub_cmd = HFC_MBSCMD_SMEINT;	/* Sub-Command */
	hfc_write_val( ap->mb->mb_init.dependent_code, 0x0000 );
	hfc_write_val( ap->mb->mb_init.type.drvlogb1.log_list_addr,
			(uint64_t)logaddr_busaddr );/* Bus address of log address list */

	/* Mailbox start/completion waiting */
	if ((rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry)) != 0) {	/* FCLNX-0523 */
		/* Error */
		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
		pci_free_consistent(ap->pci_cfginf,
			LOG_LENGTH_MEINT, (void *)trc_log_area, logout_busaddr);

		unlock_mailbox( ap );	/* Unlock mailbox */
		return ( rtn );
	}

	unlock_mailbox( ap );	/* Unlock mailbox */

	/*--------------------------------------------------------------*
	 * Data copy 
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)trc_log_area, (char *)(ulong)diag->addr, LOG_LENGTH_MEINT ) != 0 ) {

		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
		pci_free_consistent(ap->pci_cfginf,
			LOG_LENGTH_MEINT, (void *)trc_log_area, logout_busaddr);

		errlog_no = 0x14 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	pci_free_consistent(ap->pci_cfginf,
		(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
	pci_free_consistent(ap->pci_cfginf,
		LOG_LENGTH_MEINT, (void *)trc_log_area, logout_busaddr);

	HFC_EXIT("meint_log");
	return( rtn );
}


/*
 * Function:    forced_log
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Forced soft LOG)
 *
 * Arguments:   
 *  ap         - struct adap_info 
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int forced_log(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	struct	logaddr_list *loglist;			/* Log address list virtual address */
	dma_addr_t	logaddr_busaddr;			/* Log address list bus address */
	char	*trc_log_area;					/* Log buffer area virtual address */
	dma_addr_t	trclog_busaddr;				/* Log buffer area bus address */

	int		j,k,rtn;
	int		loop_cnt;

	uchar	errlog_no ;

	HFC_ENTRY("forced_log") ;

	/*--------------------------------------------------------------*
	 * Log area specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unallocated? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_FRCLOG, 0x00 );
		return( EINVAL );
	}
	if( diag->length != LOG_LENGTH_SOFT ) {	/* Is length 72K bytes ? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_FRCLOG, 0x01 );
		return( EINVAL );
	}
	/*--------------------------------------------------------------*
	 * Allocate logout list area
	 *--------------------------------------------------------------*/

	loglist = (struct logaddr_list *)pci_alloc_consistent(ap->pci_cfginf,
		(uint)sizeof(struct logaddr_list), &logaddr_busaddr);

	if (loglist == NULL) {
		errlog_no = 0x05 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x60, &errlog_no, 1) ;
		return ( ENOMEM );
	}

	BZERO((char *)loglist, (uint)sizeof(struct logaddr_list));

	/*--------------------------------------------------------------*
	 *  Allocate log buffer area
	 *--------------------------------------------------------------*/

	trc_log_area = (uchar *)pci_alloc_consistent(ap->pci_cfginf,
		LOG_LENGTH_SOFT, &trclog_busaddr);

	if (trc_log_area == NULL) {
		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);

		errlog_no = 0x08 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x60, &errlog_no, 1) ;
		return ( ENOMEM );
	}

	BZERO((char *)trc_log_area, LOG_LENGTH_SOFT);

	/*--------------------------------------------------------------*
	 * Set up log addresses to log list
	 *--------------------------------------------------------------*/
	loglist->segnum = LOG_SEGNUM_SOFT;		/* Number of segments */

	k = 0 ;
	loop_cnt = 0 ;

	if ( LOG_LENGTH_SOFT % HFC_PAGE_SIZE ) {
		/* If residue exists, loop number is (Page number + 1) */
		loop_cnt = (LOG_LENGTH_SOFT / HFC_PAGE_SIZE) + 1 ;
	}
	else {
		/* If no residue exists, loop number is page number */
		loop_cnt = LOG_LENGTH_SOFT / HFC_PAGE_SIZE ;
	}

	 /* Set bus address to log list */
	for ( j = 0 ; j < loop_cnt ; j++) {
		hfc_write_val(loglist->bus_addr[k], (uint64_t)(trclog_busaddr + HFC_PAGE_SIZE*j ) );

		if ( k >= LOG_SEGNUM_SOFT ) {	/* Set number(It must not be usually) */
			pci_free_consistent(ap->pci_cfginf,
				(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
			pci_free_consistent(ap->pci_cfginf,
				LOG_LENGTH_SOFT, (void *)trc_log_area, trclog_busaddr);

			return ( EIO );
		}

		k++ ;
	}

	/*--------------------------------------------------------------*
	 * Mailbox processing
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock of mailbox */

	/* The mailbox control block is assembled according to DRVLOGB1 */
	ap->mb-> mb_init.command = HFC_MBCMD_LOGTRACE;	/* Command */
	ap->mb-> mb_init.sub_cmd = HFC_MBSCMD_FORCSLOG;	/* Sub-Command */
	hfc_write_val( ap->mb->mb_init.dependent_code, diag->uni.forced_log.dependent_code );
	hfc_write_val( ap->mb->mb_init.type.drvlogb1.log_list_addr,
			(uint64_t)logaddr_busaddr );/* Bus address of log address list */

	/* Mailbox start/completion waiting */
	if ((rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry)) != 0) {	/* FCLNX-0524 */
		/* Error case */
		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
		pci_free_consistent(ap->pci_cfginf,
			LOG_LENGTH_SOFT, (void *)trc_log_area, trclog_busaddr);

		unlock_mailbox( ap );	/* Unlock mailbox */
		return ( rtn );
	}

	unlock_mailbox( ap );	/*  Unlock mailbox */


	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)trc_log_area, (char *)(ulong)diag->addr, LOG_LENGTH_SOFT ) != 0 ) {

		pci_free_consistent(ap->pci_cfginf,
			(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
		pci_free_consistent(ap->pci_cfginf,
			LOG_LENGTH_SOFT, (void *)trc_log_area, trclog_busaddr);

		errlog_no = 0x0d ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	pci_free_consistent(ap->pci_cfginf,
		(uint)sizeof(struct logaddr_list), (void *)loglist, logaddr_busaddr);
	pci_free_consistent(ap->pci_cfginf,
		LOG_LENGTH_SOFT, (void *)trc_log_area, trclog_busaddr);

	HFC_EXIT("forced_log");

	return( rtn );
}


/*
 * Function:    fw_post
 *
 * Purpose:     Processing of subcommand (F/W POST) of ioctl(HFCDIAG0)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int fw_post(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int		rtn;

	HFC_ENTRY("fw_post");

	/*--------------------------------------------------------------*
	 * Parameter check (command check)
	 *--------------------------------------------------------------*/
	if( (diag->uni.post.cmd == POST_CMD_IMEM)	||		/* Internal RAM test */
	    (diag->uni.post.cmd == POST_CMD_ILOOP)	||		/* Internal LOOP test */
	    (diag->uni.post.cmd == POST_CMD_DMA)	||		/* DMA test */
	    (diag->uni.post.cmd == POST_CMD_ELOOP)  ||		/* External LOOP test */
	    (diag->uni.post.cmd == POST_CMD_BOOT)  ||		/* BOOT test */
	    (diag->uni.post.cmd == POST_CMD_BOOT_MLT)  ||	/* Multi BOOT test */
	    (diag->uni.post.cmd == POST_CMD_ILOOP_MLT)  ||	/* Multi port internal LOOP test */
	    (diag->uni.post.cmd == POST_CMD_ELOOP_MLT)) {	/* Multi port external LOOP test */
	}
	else {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_FWPOST, 0x00 );
		return ( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Mailbox processing
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock mailbox */

	/* The mailbox control block is assembled according to DRVLOGB0 */
	ap->mb->mb_init.command = HFC_MBCMD_LOGTRACE;				/* Command */
	ap->mb->mb_init.sub_cmd = HFC_MBSCMD_DIAG;					/* Sub-Command */
	hfc_write_val( ap->mb->mb_init.dependent_code, (ushort)diag->uni.post.cmd );/* Test number */

	/* Set up mailbox control block  */
	if (( rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry) ) == 0 ){
		/* Copy RSLT information to internal DIAG area */
		diag->uni.post.reslt = ap->mb->mb_resp.type.drvlogb0.diag.reslt;
		/* Set error code */
		diag->uni.post.err_code[0] = ap->mb->mb_resp.type.drvlogb0.diag.resv0[0];
		diag->uni.post.err_code[1] = ap->mb->mb_resp.type.drvlogb0.diag.resv0[1];
		diag->uni.post.err_code[2] = ap->mb->mb_resp.type.drvlogb0.diag.resv0[2];
		/* IMLLOG output (POST ERROR) */
		if ( ( diag->uni.post.reslt == 0xff )||( (diag->uni.post.reslt & 0xf0) == 0x90 ) ){  
			hfc_logout(ap, (uint)0x77, HFC_ERRLOG_TYPE_IMLLOG);
			hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_IMLLOG, ERRID_HFCP_ERRC, 0x77, NULL, 0);
		}
	}

	unlock_mailbox( ap );	/* Lock release of mailbox */

	/* Output IMLLOG (Timeout case) */
	if (rtn == ETIMEDOUT ){
		hfc_logout(ap, (uint)0x78, HFC_ERRLOG_TYPE_IMLLOG);
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_IMLLOG, ERRID_HFCP_ERRC, 0x78, NULL, 0);
	}

	HFC_EXIT("fw_post") ;

	return( rtn );
}


/*
 * Function:    fw_start
 *
 * Purpose:     Processing for ioctl(HFCDIAG0) subcommand (F/W Start) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int fw_start(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uint i , lp ;
	int rc=0;
	int	lun_no;
	ulong	flags = 0 ;
	struct adap_info 	*ap_work ;
	struct target_info 	*target_work ;
	struct	mp_adap_info	*mpap; 	/* struct mp_adap_info */
	
	HFC_ENTRY("fw_start");

	mpap = ap->mp_adap_info;

	/*--------------------------------------------------------------*
	 * Parameter check (flag)
	 *--------------------------------------------------------------*/
	if ( ap->pkg.type == HFC_PKTYPE_FPP )
	{
		if( (diag->uni.fw_start.mode != FW_MODE_NORMAL)		&&
		    (diag->uni.fw_start.mode != FW_MODE_DIAG)		&&
	    	(diag->uni.fw_start.mode != FW_MODE_NAI_LOOP)	&&
	    	(diag->uni.fw_start.mode != FW_MODE_GAI_LOOP) )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_FWSTART, 0x00 );
			return ( EINVAL );
		}
	}
	if ( ap->pkg.type == HFC_PKTYPE_FIVE )
	{
		if( (diag->uni.fw_start.mode != FW_MODE_NORMAL)	&&
		    (diag->uni.fw_start.mode != FW_MODE_DIAG) )	
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_FWSTART, 0x01 );
			return ( EINVAL );
		}
	}
	if( ap->fcp_mode != FW_MODE_NORMAL )
	{
		if( diag->uni.fw_start.mode != FW_MODE_NORMAL )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_FWSTART, 0x02 );
			return ( EINVAL );
		}
	}
	if( diag->uni.fw_start.mode != FW_MODE_NORMAL )
	{

		HFC_ADAP_LOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
		if( (test_bit(HFC_DIAG_PROGRESS, (ulong *)&ap->mp_adap_info->status))||
			(test_bit(HFC_HMCK_RECOVRTY, (ulong *)&ap->mp_adap_info->lock)) )
		{

			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_FWSTART, 0x03 );
			HFC_ADAP_UNLOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
			return ( EBUSY ) ;
		}
		ap_work = ap->mp_adap_info->ap ;
		while( ap_work != NULL )
		{
			spin_lock_irqsave(&ap_work->adap_lock,flags) ;
			for(lp=0 ; lp<MAX_TARGET_PROBE ; lp++)
			{
				target_work = ap_work->target_arg[lp] ;
				while( target_work != NULL )
				{
					if( !test_bit(HFC_TARGETINF_VALID,(ulong *)&target_work->flags) )
					{
						target_work = target_work->next ;
						continue ;
					}
					if( (target_work->we_que_cnt != 0) || (target_work->wx_que_cnt != 0) )
					{
						HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
											HFC_TRC_IOCTL_FWSTART, 0x04 );
						HFC_ADAP_UNLOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
						spin_unlock_irqrestore( &ap_work->adap_lock,flags ) ;
						return ( EBUSY ) ;
					}
					target_work = target_work->next ;
				}
			}

			while( test_bit( HFC_MAILBOX_BUSY, (ulong *)&( ap_work -> mb_lock ) ) )
			{
				set_bit(HFC_DIAG_DELAY, (ulong *)&ap_work->status );
				/* 3s wait during processing mailbox */
				hfc_w_stop(ap_work,HFC_DIAG_DELAY_TMR) ;   
				hfc_w_start(ap_work,HFC_DIAG_DELAY_TMR) ;
				spin_unlock_irqrestore(&ap_work->adap_lock,flags) ;

				hfc_sleep_on(&ap->mb_event, &ap->mb_event_wait);					/* FCLNX-0269 */
				clear_bit(HFC_DIAG_DELAY, (ulong *)&ap_work->status );
				spin_lock_irqsave(&ap_work->adap_lock,flags) ;
			}

			clear_bit( HFC_ONLINE, (ulong *)&ap_work -> status );

			spin_unlock_irqrestore(&ap_work->adap_lock,flags) ;
			ap_work = ap_work->next ;
		}
		set_bit(HFC_DIAG_PROGRESS , (ulong *)&ap->mp_adap_info->status) ;
		HFC_ADAP_UNLOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
	}

	switch (diag->uni.fw_start.mode)
	{
		case  FW_MODE_DIAG :
		case  FW_MODE_NAI_LOOP :	
		case  FW_MODE_GAI_LOOP :	
			/* Initiate boot microprogram and check POST result */
			if ( ap->pkg.type == HFC_PKTYPE_FPP )
			{
				mdelay(10);

				/* Set flag in INI_RST */									/* FCLNX-282 STR */
				HFC_ADAP_LOCK(mpap,HFC_MP_ADAP_BUSY);
				set_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );		/* FCLNX-282 END */

				/* Start STOP_FUNCTION */
				hfc_write_reg(ap, HFC_IOSPACE_CMDCTL, PCI_LENGTH_01, 0x50);
				/* Wait 10ms */
				mdelay(10);

				hfc_reset_start(ap,HFC_INI_RESET);
				mdelay(5); 	/* Wait 5ms */

				/* Reset flag in INI_RST */									/* FCLNX-282 STR*/
				clear_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );
				HFC_ADAP_UNLOCK(mpap,HFC_MP_ADAP_BUSY);						/* FCLNX-282 END*/

				clear_bit( HFC_MAILBOX_BUSY, (ulong *)&( ap -> mb_lock ) );
				hfc_reset_start(ap,HFC_WSCA_CLEAR);

				clear_bit(HFC_CHK_STOP , (ulong *)&ap->mp_adap_info->status) ;

				hfc_reset_start(ap,HFC_REBOOT);

				/* 500ms wait */
				i = 0;
				mdelay(500);
				
				/* Check POST result (no retry) */
				rc = hfc_config_hw_set(ap,0);
			}
			else if ( ap->pkg.type == HFC_PKTYPE_FIVE )
			{
				mdelay(10);

				if( ap->diag_cnt == 0 )											/* FCLNX-293 STR*/
				{
					/* Set flag in CTL_RST */									/* FCLNX-282 STR*/
					HFC_ADAP_LOCK(mpap,HFC_MP_ADAP_BUSY);
					set_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );		/* FCLNX-282 END*/

					/* Start STOP_FUNCTION */
					hfc_write_reg(ap, HFC_IOSPACE_CMDCTL, PCI_LENGTH_01, 0x50);
					/* Wait 10ms */
					mdelay(10);

					hfc_reset_start(ap, HFC_CTLRST);
					mdelay(10); 	/* Wait 10ms */

					/* Reset flag in CTL_RST */									/* FCLNX-282 STR*/
					clear_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );
					HFC_ADAP_UNLOCK(mpap,HFC_MP_ADAP_BUSY);						/* FCLNX-282 END*/

					ap_work = ap->mp_adap_info->ap ;
					while( ap_work != NULL )
					{
						clear_bit( HFC_MAILBOX_BUSY, (ulong *)&( ap_work -> mb_lock ) );
						ap_work = ap_work->next;
					}

					hfc_reset_start(ap,HFC_WSCA_CLEAR);

					clear_bit(HFC_CHK_STOP , (ulong *)&ap->mp_adap_info->status) ;

					hfc_reset_start(ap,HFC_REBOOT);

					/* Wait 2s */
					i = 0;

					set_bit(HFC_DIAG_DELAY, (ulong *)&ap->status );
					/* Wait 2s */
					hfc_w_stop(ap,HFC_DIAG_DELAY_TMR) ;
					hfc_w_start(ap,HFC_DIAG_DELAY_TMR) ;
					/* Wait timeout for completion of DIAG process */
					hfc_sleep_on(&ap->mb_event, &ap->mb_event_wait);				/* FCLNX-0296 */
					clear_bit(HFC_DIAG_DELAY, (ulong *)&ap->status );
					
					/* Check POST result (no retry) */
					rc = hfc_config_hw_set_five(ap,0);
				}
				else{
					rc = 0;
				}
				ap->diag_cnt++;													/* FCLNX-293 END*/
			}

			/* POST result check */
			if( rc != 0 ) {
				HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
						HFC_TRC_IOCTL_FWSTART, 0x01 );
				return ( EIO );
			}
			else
			{
				hfc_reset_start(ap,HFC_SET_INIADR);			/* INITTBL ADR(0x310)  */
															/* ALPA(0x319)         */
				hfc_write_reg(ap, HFC_IOSPACE_CA_FLAG, PCI_LENGTH_01, diag->uni.fw_start.mode);
				hfc_write_reg(ap, HFC_IOSPACE_FRAMEA, PCI_LENGTH_04, HFC_FRAMEA_FW_START);

				/* Reconfigure interrupt mask */
				hfc_write_reg( ap, HFC_IOSPACE_INTA_MSK,PCI_LENGTH_04, HFC_INTA_MASK );
			}
			if( (diag->uni.fw_start.mode == FW_MODE_NAI_LOOP)	||
			    (diag->uni.fw_start.mode == FW_MODE_GAI_LOOP) )
			{
				for ( lun_no = 0 ; lun_no < MAX_LOOP_LUN ; lun_no++ )
				{
					ap->loop_dev_info[lun_no].flag |= HFC_LOOP_EXEC;		/* LOOP test start */
				}
			}
			break ;

		case  FW_MODE_NORMAL :
			if (diag->uni.fw_start.err_code == FW_DIAG_FAILED)
			{
				spin_lock_irqsave( &ap->adap_lock,flags );
				/* Check stop if DIAG fails */
				HFC_ISSUE_CSTP( ap, FALSE ) ;									/* FCLNX-0279 *//* @MLPF */
				/* Return to previous interruption level */
				spin_unlock_irqrestore( &ap->adap_lock,flags );

				/* Clear error code */
				diag->uni.fw_start.err_code = 0x00 ;

				return( 0 );
			}
			else
			{
				if ( ap->pkg.type == HFC_PKTYPE_FPP )
				{
					
					ap_work = ap->mp_adap_info->ap ;
					while( ap_work != NULL )
					{
						hfc_reset_adap_info(ap_work) ;
						ap_work = ap_work->next ;
					}

					/* Set flag in INI_RST */									/* FCLNX-282 STR*/
					HFC_ADAP_LOCK(mpap,HFC_MP_ADAP_BUSY);
					set_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );		/* FCLNX-282 END*/

					/* Initiate STOP_FUNCTION  */
					hfc_write_reg(ap, HFC_IOSPACE_CMDCTL, PCI_LENGTH_01, 0x50);
					/* 1ms wait */
					mdelay(1);

					hfc_reset_start(ap,HFC_INI_RESET);
					hfc_reset_start(ap,HFC_WSCA_CLEAR);
					hfc_reset_start(ap,HFC_SET_WS04);			/* WS Comu Area(0x318) */
					hfc_reset_start(ap,HFC_SET_INIADR);			/* INITTBL ADR(0x310)  */
																/* ALPA(0x319)         */
					/* Reset flag in INI_RST */									/* FCLNX-282 STR*/
					clear_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );
					HFC_ADAP_UNLOCK(mpap,HFC_MP_ADAP_BUSY);						/* FCLNX-282 END*/

					hfc_reset_start(ap,HFC_REBOOT);
				
					/* Wait 500ms */
					mdelay(500);	/* FCLNX_0012 */
		
					/* Check POST result (no retry) */
					if( hfc_config_hw_set(ap,0) )
					{
						/*-- Init fail --*/
						HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
								HFC_TRC_IOCTL_FWSTART, 0x01 );
						return ( EIO );
					}
					else
					{
						HFC_ADAP_LOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
						clear_bit(HFC_DIAG_PROGRESS , (ulong *)&ap->mp_adap_info->status);
						HFC_ADAP_UNLOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
						ap_work = ap->mp_adap_info->ap ;
						while( ap_work != NULL )
						{
							set_bit(HFC_WAIT_LINKUP,(ulong *)&ap_work->status) ;
							set_bit(HFC_DIAG_END,(ulong *)&ap_work->status) ;
							hfc_inta_mask_set(ap_work,HFC_INTA_MASK) ;
							ap_work = ap_work->next ;
						}
					}

					for ( lun_no = 0 ; lun_no < MAX_LOOP_LUN ; lun_no++ ){
						if( ap->loop_dev_info[lun_no].flag & HFC_LOOP_EXEC) {
							ap -> loop_dev_info[lun_no].flag &= ~(HFC_LOOP_EXEC);	/* LOOP test end */
						}
					}
				}
				else if ( ap->pkg.type == HFC_PKTYPE_FIVE )
				{
					if( test_bit(HFC_HWCHKSTOP, (ulong *)&ap->mp_adap_info->status ) )
					{
						return(0);
					}

					/* Set flag in CTL_RST */										/* FCLNX-282 STR*/
					HFC_ADAP_LOCK(mpap,HFC_MP_ADAP_BUSY);
					set_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );			/* FCLNX-282 END*/

					/* Initiate STOP_FUNCTION */
					hfc_write_reg(ap, HFC_IOSPACE_CMDCTL, PCI_LENGTH_01, 0x50);
					mdelay(1); 	/* 1ms wait */

					hfc_reset_start(ap, HFC_CTLRST);

					mdelay(1); 	/* 1ms wait */

					/* Reset flag in CTL_RST */										/* FCLNX-282 STR*/
					clear_bit( HFC_CTLRST_PROCESS, (ulong *)&mpap->status );
					HFC_ADAP_UNLOCK(mpap,HFC_MP_ADAP_BUSY);							/* FCLNX-282 END*/

					hfc_reset_start(ap, HFC_WSCA_CLEAR);
					hfc_reset_adap_info(ap) ;
					hfc_reset_start(ap, HFC_REBOOT);

					set_bit(HFC_DIAG_DELAY, (ulong *)&ap->status );
					/* 2s wait */
					hfc_w_stop(ap,HFC_DIAG_DELAY_TMR) ;
					hfc_w_start(ap,HFC_DIAG_DELAY_TMR) ;
					/* Wait timeout for completion of DIAG process */
					hfc_sleep_on(&ap->mb_event, &ap->mb_event_wait );				/* FCLNX-0269 */

					clear_bit(HFC_DIAG_DELAY, (ulong *)&ap->status );
					
					if( hfc_config_hw_set_five(ap,0) )
					{
						/*-- init fail --*/
						HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
								HFC_TRC_IOCTL_FWSTART, 0x21 );
						return ( EIO );
					}
					else
					{
						HFC_ADAP_LOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
						clear_bit(HFC_DIAG_PROGRESS , (ulong *)&ap->mp_adap_info->status);
						HFC_ADAP_UNLOCK(ap->mp_adap_info , HFC_MP_ADAP_BUSY) ;
						/* WS Comu Area(0x318) */
						hfc_reset_start(ap, HFC_SET_WS80);
						/* INITTBL ADR(0x310)  */
						/* ALPA(0x319)         */
						hfc_reset_start(ap, HFC_SET_INIADR);
						
						hfc_reset_start(ap, HFC_FW_START);
						set_bit( HFC_WAIT_LINKUP, (ulong *)&ap->status );
						set_bit( HFC_DIAG_END, (ulong *)&ap->status );
						/* Set interrupt mask */
						hfc_write_reg(ap,( uint )HFC_IOSPACE_INTA_MSK,( char )0x4, HFC_INTA_MASK);
					}
					ap_work = ap->mp_adap_info->ap ;
					while( ap_work != NULL )
					{
						if( ap_work != ap )
							break;
						
						ap_work = ap_work->next ;
					}
					if( ap_work != NULL )
					{
						/* WS Comu Area(0x318) */
						hfc_reset_start(ap_work, HFC_SET_WS80);
						/* INITTBL ADR(0x310)  */
						/* ALPA(0x319)         */
						hfc_reset_start(ap_work, HFC_SET_INIADR);
						hfc_reset_start(ap_work, HFC_FW_START);                      /* FCWIN-0237 */
						set_bit( HFC_WAIT_LINKUP, (ulong *)&ap_work->status );
						set_bit( HFC_DIAG_END, (ulong *)&ap_work->status );
						/* Set interrupt mask */
						hfc_write_reg(ap_work,( uint )HFC_IOSPACE_INTA_MSK,( char )0x4, HFC_INTA_MASK);
					}
				}
				set_bit(HFC_DIAG_DELAY, (ulong *)&ap->status );
				hfc_w_stop(ap,HFC_DIAG_DELAY_TMR) ;
				hfc_w_start(ap,HFC_DIAG_DELAY_TMR) ;
				/* Wait timeout for completion of DIAG process */
				hfc_sleep_on(&ap->mb_event, &ap->mb_event_wait );					/* FCLNX-0269 */
				clear_bit(HFC_DIAG_DELAY, (ulong *)&ap->status );
			}
			break ;

	}	/*-- end of switch --*/
	HFC_EXIT("fw_start");

	return( 0 );
}


/*
 * Function:    stop_func
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Stop_Function) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *             - 0
 * Notes:       
 */
int stop_func(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{

	HFC_ENTRY("stop_func") ;

	/* Initiste STOP_FUNCTION  */
	hfc_write_reg(ap, HFC_IOSPACE_CMDCTL, PCI_LENGTH_01, 0x40);

	/* Drop ONLINE bit to prevent from initiating SCSI and ioctl */
	clear_bit(HFC_ONLINE,(ulong *)&ap->status) ;

	HFC_EXIT("stop_func") ;
	return( 0 );
}


																					/* FIVE STR */
/*
 * Function:    init_mode_set
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (INIT MODE SET)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl 
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int init_mode_set(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int rtn = 0;

	HFC_ENTRY("init_mode_set") ;

	/*--------------------------------------------------------------*
	 * Lock Mailbox 
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock of mailbox */

	/*--------------------------------------------------------------*
	 * Mailbox processing
	 *--------------------------------------------------------------*/
	/* Setup mailbox control block */
	ap->mb->mb_init.command = HFC_MBCMD_FPPCTL;						/* Command */
	ap->mb->mb_init.sub_cmd = HFC_MBSCMD_INITMDSET;					/* Sub-Command */
	hfc_write_val( ap->mb->mb_init.dependent_code, 0x0000 );		/* Test number */

	ap->mb->mb_init.type.drvctl.un.init_mode_set.mode[0] = diag->uni.init_mode_set.mode_alpa.uc_mode[3];
	ap->mb->mb_init.type.drvctl.un.init_mode_set.mode[1] = diag->uni.init_mode_set.mode_alpa.uc_mode[2];
	ap->mb->mb_init.type.drvctl.un.init_mode_set.mode[2] = diag->uni.init_mode_set.mode_alpa.uc_mode[1];
	ap->mb->mb_init.type.drvctl.un.init_mode_set.al_pa = 0;
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[0] = (uchar)'H';
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[1] = (uchar)'I';
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[2] = (uchar)'T';
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[3] = (uchar)'A';
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[4] = (uchar)'C';
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[5] = (uchar)'H';
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[6] = (uchar)'I';
	ap->mb->mb_init.type.drvctl.un.init_mode_set.magic_num[7] = (uchar)' ';
	if( ( ( diag->uni.init_mode_set.mode_alpa.mode >> 8 ) & 0x00FFFFFF ) != HFC_INITMDSET_FCP ){
		ap->mb->mb_init.type.drvctl.un.init_mode_set.al_pa = diag->uni.init_mode_set.mode_alpa.al_pa[0];
		hfc_write_val( ap->mb->mb_init.type.drvctl.un.init_mode_set.connect_type, diag->uni.init_mode_set.connect_type );
		hfc_write_val( ap->mb->mb_init.type.drvctl.un.init_mode_set.wwpn, diag->uni.init_mode_set.wwpn );
		hfc_write_val( ap->mb->mb_init.type.drvctl.un.init_mode_set.wwnn, diag->uni.init_mode_set.wwnn );

		hfc_write_val( ap->mb->mb_init.type.drvctl.un.init_mode_set.buf_size, diag->uni.init_mode_set.buf_size );
		hfc_write_val( ap->mb->mb_init.type.drvctl.un.init_mode_set.xrdy_div, diag->uni.init_mode_set.xrdy_div );
		hfc_write_val( ap->mb->mb_init.type.drvctl.un.init_mode_set.param, diag->uni.init_mode_set.param );
	}

	/* Start Mailbox process and wait for process completion */
	rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry);	/* FCLNX-0523 */

	unlock_mailbox( ap );	/* Unlock mailbox */

	HFC_EXIT("init_mode_set") ;
	return( rtn );

}	/* end of init_mode_set */


/*
 * Function:    fcp_mode_set
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (FCP MODE SET)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  ETIMEDOUT  - Mailbox initiation time out
 *  EIO        - Other errors
 * Notes:       
 */
int fcp_mode_set(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int rtn = 0;

	HFC_ENTRY("fcp_mode_set") ;

	/*--------------------------------------------------------------*
	 * Lock Mailbox 
	 *--------------------------------------------------------------*/
	lock_mailbox( ap );		/* Lock of mailbox */

	/*--------------------------------------------------------------*
	 * Mailbox processing 
	 *--------------------------------------------------------------*/
	/* Setup mailbox control block */
	ap->mb->mb_init.command = HFC_MBCMD_FPPCTL;						/* Command */
	ap->mb->mb_init.sub_cmd = HFC_MBSCMD_FCPMDSET;					/* Sub-Command */
	hfc_write_val( ap->mb->mb_init.dependent_code, 0x0000 );		/* Test no. */

	hfc_write_val( ap->mb->mb_init.type.drvctl.un.fcp_mode_set.act_ctl, diag->uni.fcp_mode_set.act_ctl );
	hfc_write_val( ap->mb->mb_init.type.drvctl.un.fcp_mode_set.il_ctl, diag->uni.fcp_mode_set.il_ctl );
	hfc_write_val( ap->mb->mb_init.type.drvctl.un.fcp_mode_set.seq_ctl, diag->uni.fcp_mode_set.seq_ctl );
	hfc_write_val( ap->mb->mb_init.type.drvctl.un.fcp_mode_set.rsp_delay, diag->uni.fcp_mode_set.rsp_delay );

	/* Start mailbox process and wait for completion */
	rtn = hfc_mailbox_proc(ap, HFC_MB_TMR, HFC_MB_TO, ap->els_retry);	/* FCLNX-0523 */

	unlock_mailbox( ap );	/* Unlock mailbox */

	HFC_EXIT("fcp_mode_set") ;
	return( rtn );

}	/* end of fcp_mode_set */
																					/* FIVE END */


/*==================================================================*
 * pci_access
 *
 *    func    : Process for ioctl(HFCDIAG0) subcommand (PCI Read)
 *            : Process for ioctl(HFCDIAG0) subcommand (PCI Write) 
 *
 *==================================================================*/
int pci_access(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	uint	wk_data;			/* Internal data area */
	ushort	wk2 ;
	uint	wk4 ;

	HFC_ENTRY("pci_access");

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	/* Virtual address check */
	if( !diag->addr ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_PCIACC, 0x00 );
		return( EINVAL );
	}
	/* PCI memory space address check */
	if ( ap->pkg.type == HFC_PKTYPE_FPP )
	{
		if( diag->uni.pci.addr > PCI_ADDR_FPP_MAX ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_PCIACC, 0x01 );
			return( EINVAL );
		}
	}
	else if ( ap->pkg.type == HFC_PKTYPE_FIVE )
	{
		if( diag->uni.pci.addr > PCI_ADDR_FIVE_MAX ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_PCIACC, 0x01 );
			return( EINVAL );
		}
	}
	/* Access length check */
	if( (diag->length == PCI_LENGTH_01) ||
	    (diag->length == PCI_LENGTH_02) ||
	    (diag->length == PCI_LENGTH_04) ) {
	}
	else {	/* Invalid access length check */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_PCIACC, 0x02 );
		return( EINVAL );
	}
	/* Access over check */
	if ( ap->pkg.type == HFC_PKTYPE_FPP )
	{
		if( (diag->uni.pci.addr + diag->length) > PCI_ADDR_FPP_MAX+1 ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_PCIACC, 0x03 );
			return( EINVAL );
		}
	}
	else if ( ap->pkg.type == HFC_PKTYPE_FIVE )
	{
		if( (diag->uni.pci.addr + diag->length) > PCI_ADDR_FIVE_MAX+1 ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_PCIACC, 0x03 );
			return( EINVAL );
		}
	}

	if( diag->sub_cmd == HFC_DG_RDPCI ) {
	/*--------------------------------------------------------------*
	 * PCI Read
	 *--------------------------------------------------------------*/

		/* PCI memory read */
		wk_data = hfc_read_reg_ext(ap, diag->uni.pci.addr, diag->length) ;
		switch( diag->length )
		{
			case PCI_LENGTH_04: wk4 = (uint)   wk_data; wk_data = hfc_read_val(wk4); break;
			case PCI_LENGTH_02: wk2 = (ushort) wk_data; wk_data = hfc_read_val(wk2); break;
		}
		/* Copy read data to user area (wk_datad -> iag->addr) */
		if ( COPYOUT( (char *)&wk_data, (char *)(ulong)diag->addr, diag->length ) != 0 ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_PCIACC, 0x04 );
			errlog_no = 0x2a ;
			hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
			return ( EFAULT );
		}
	}
	else {
	/*--------------------------------------------------------------*
	 * PCI Write
	 *--------------------------------------------------------------*/

		/* Copy write data to internal area (diag->addr -> wk_data) */
		wk_data = 0 ;
		if ( COPYIN( (char *)(ulong)diag->addr, (char *)&wk_data, diag->length )  != 0 ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_PCIACC, 0x05 );
			errlog_no = 0x2b ;
			hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5E, &errlog_no, 1) ;
			return ( EFAULT );
		}
		switch( diag->length )
		{
			case PCI_LENGTH_04: wk4 = (uint)   wk_data; wk_data = hfc_read_val(wk4); break;
			case PCI_LENGTH_02: wk2 = (ushort) wk_data; wk_data = hfc_read_val(wk2); break;
		}
		/* PCI memory write */
		hfc_write_reg_ext(ap, diag->uni.pci.addr, diag->length, wk_data);
	}

	HFC_EXIT("pci_access");
	return( 0 );
}


/*==================================================================*
 * pci_cnf_access
 *
 *    func    : Process for ioctl(HFCDIAG0) subcommand (PCI Config Read) 
 *            : Process for ioctl(HFCDIAG0) subcommand (PCI Config Write) 
 *
 *==================================================================*/
int pci_cnf_access(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	uint	wk_data;			/* Internal data area */
	int		rtn = 0;

	HFC_ENTRY("pci_cnf_access" ) ;

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	/* Virtual address check */
	if( !diag->addr )
	{
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_PCICNFACC, 0x00 );
		return( EINVAL );
	}
	/* PCI config space address check */
	if( diag->uni.pci.addr > PCI_CNF_ADDR_MAX ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_PCICNFACC, 0x01 );
		return( EINVAL );
	}
	/* Access length check */
	if( (diag->length == PCI_LENGTH_01) ||
	    (diag->length == PCI_LENGTH_02) ||
	    (diag->length == PCI_LENGTH_04) ) {
	}
	else {	/* Invalid access length check */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_PCICNFACC, 0x02 );
		return( EINVAL );
	}
	/* Access over check */
	if( (diag->uni.pci.addr + diag->length) > PCI_CNF_ADDR_MAX+1 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
			HFC_TRC_IOCTL_PCICNFACC, 0x03 );
		return( EINVAL );
	}

	if( diag->sub_cmd == HFC_DG_RDPCI_CFG ) {
	/*--------------------------------------------------------------*
	 * PCI Config Read
	 *--------------------------------------------------------------*/
		/* PCI config reading */
		wk_data = hfc_read_cnfg(ap, diag->uni.pci.addr, diag->length);
		wk_data = hfc_read_val(wk_data) ;
		if( diag->length == PCI_LENGTH_01 ) {
			wk_data <<= 24;
		}
		else if( diag->length == PCI_LENGTH_02 ) {
			wk_data <<= 16;
		}
		/* Copy read data to user area (wk_datad -> iag->addr) */
		if ( COPYOUT( (char *)&wk_data, (char *)(ulong)diag->addr, diag->length ) != 0 ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_PCICNFACC, 0x04 );
			errlog_no = 0x2c ;
			hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
			return ( EFAULT );
		}
	}
	else {
	/*--------------------------------------------------------------*
	 * PCI Config Write
	 *--------------------------------------------------------------*/
		/* Copy write data to internal area (diag->addr -> wk_data) */
		if ( COPYIN( (char *)(ulong)diag->addr, (char *)&wk_data, diag->length )  != 0 ) {
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_PCICNFACC, 0x05 );
			errlog_no = 0x2d ;
			hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5E, &errlog_no, 1) ;
			return ( EFAULT );
		}
		wk_data = hfc_read_val(wk_data) ;
		/* Shift write data */
		if( diag->length == PCI_LENGTH_01 ) {
			wk_data >>= 24;
		}
		else if( diag->length == PCI_LENGTH_02 ) {
			wk_data >>= 16;
		}
		/* PCI config write */
		hfc_write_cnfg(ap, diag->uni.pci.addr, diag->length, wk_data);
	}
	HFC_EXIT("pci_cnf_access" ) ;

	return( rtn );
}


/*
 * Function:    read_fw_init_tbl
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read fw_init_tbl) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    - 0
 *
 * Notes:       
 */
int read_fw_init_tbl(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_fw_init_tbl");

	/*--------------------------------------------------------------*
	 * Log area specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_FWINIT, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct fw_init_tbl) ) {	/* Is data size short for fw_init_tbl ?*/
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_FWINIT, 0x01 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)ap->fw_init_p, (char *)(ulong)diag->addr, sizeof(struct fw_init_tbl) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_FWINIT, 0x02 );
		errlog_no = 0x15 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_fw_init_tbl");

	return( 0 );
}


/*
 * Function:    read_xob
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read xob) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    - 0
 *
 * Notes:       
 */
int read_xob(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_xob") ;

	/*--------------------------------------------------------------*
	 * Specificaton check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XOB, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct xob) ) {	/* Is data size short for xob ?*/
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XOB, 0x01 );
		return( EINVAL );
	}
	if( 
		(diag->uni.xob_xrb_scmd.no >= hfc_read_val(ap->fw_init_p->xob_num)) ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XOB, 0x02 );
		return( EINVAL );
	}
	if( diag->uni.xob_xrb_scmd.cnt != MAX_READ_CNT ) {	/* Read number  */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XOB, 0x03 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)(ap->xob + diag->uni.xob_xrb_scmd.no),
		(char *)(ulong)diag->addr, sizeof(struct xob) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XOB, 0x04 );
		errlog_no = 0x16 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_xob");
	return( 0 );
}


/*
 * Function:    read_xrb
 *
 * Purpose:     Processing for ioctl(HFCDIAG0) subcommand (Read xrb) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    - 0
 *           
 * Notes:       
 */
int read_xrb(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_xrb");

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {		/* Is virtual address unassigned ? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XRB, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct xrb) ) {	/* Is size short for xrb? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XRB, 0x01 );
		return( EINVAL );
	}
	if( (diag->uni.xob_xrb_scmd.no >= hfc_read_val(ap->fw_init_p->xrb_num)) ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XRB, 0x02 );
		return( EINVAL );
	}
	if( diag->uni.xob_xrb_scmd.cnt != MAX_READ_CNT ) {	/* Read number */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XRB, 0x03 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)(ap->xrb + diag->uni.xob_xrb_scmd.no),
		(char *)(ulong)diag->addr, sizeof(struct xrb) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_XRB, 0x04 );
		errlog_no = 0x17 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_xrb");

	return( 0 );
}


/*
 * Function:    read_hfcpkt
 *
 * Purpose:     Process for ioctl(HFCDIAG0)subcommand (Read hfc_pkt)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     0
 *
 * Notes:       2004.11.30
 */
int read_hfcpkt(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;
	char *ptr;

	HFC_ENTRY("read_hfc_pkt");

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCPKT, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct hfc_pkt) ) {	/* Is size short for hfc_pkt? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCPKT, 0x01 );
		return( EINVAL );
	}
	if( diag->uni.xob_xrb_scmd.no >= ap->pkt_num ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCPKT, 0x02 );
		return( EINVAL );
	}
	if( diag->uni.xob_xrb_scmd.cnt != MAX_READ_CNT ) {	/* Read number */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCPKT, 0x03 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	ptr = (char *)&ap->pkt_pool[ap->pkt_no / HFC_PKT_POOL_SIZE][ap->pkt_no % HFC_PKT_POOL_SIZE];	/* FCLNX-0521 */
	if ( COPYOUT( ptr, (char *)(ulong)diag->addr, sizeof(struct hfc_pkt) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCPKT, 0x04 );
		errlog_no = 0x17 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_hfcpkt");

	return( 0 );
}


/*
 * Function:    read_seg_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read seg_info) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *             0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 * Notes:       
 */
int read_seg_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int seg_cnt;				/* Number of segments */
	int seg_info_cnt;			/* DMA segment info array */

	uchar	errlog_no ;

	HFC_ENTRY("read_seg_info");

	seg_cnt = (ap->dma_max / HFC_PAGE_SIZE);
	seg_info_cnt = seg_cnt + seg_cnt / ( HFC_PAGE_SIZE / sizeof( struct seg_info ) );

	/*--------------------------------------------------------------*
	 * Specified check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {		/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_SEG, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct seg_info) ) {	/* Is size short for seg_info? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_SEG, 0x01 );
		return( EINVAL );
	}
	if( (diag->uni.xob_xrb_scmd.no >= seg_info_cnt) ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_SEG, 0x02 );
		return( EINVAL );
	}
	if( diag->uni.xob_xrb_scmd.cnt != MAX_READ_CNT ) {	/* Read number  */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_SEG, 0x03 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)(ap->seg_info + diag->uni.xob_xrb_scmd.no),
		(char *)(ulong)diag->addr, sizeof(struct seg_info) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_SEG, 0x04 );
		errlog_no = 0x18 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_seg_info");
	return( 0 );
}


																					/* FIVE STR */
/*
 * Function:    read_mpadap_info
 *
 * Purpose:     Processing of subcommand (Read hfc_mpadap_info) of ioctl(HFCDIAG0)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 * Notes:       
 */
int read_mpadap_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_mpadap_info");
	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MPADAP, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct mp_adap_info) ) {	/* Is size short for mp_adap_info? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MPADAP, 0x01 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)ap->mp_adap_info, (char *)(ulong)diag->addr, sizeof(struct mp_adap_info) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MPADAP, 0x02 );
		errlog_no = 0x1d ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_mpadap_info") ;
	return( 0 );
}	/* end of read_mpadap_info */
																					/* FIVE END */


/*
 * Function:    read_adap_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read adap_info)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 * Notes:       
 */
int read_adap_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_adap_info") ;

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_APINFO, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct adap_info) ) {	/* Is size short for adap_info? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_APINFO, 0x01 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)ap, (char *)(ulong)diag->addr, sizeof(struct adap_info) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_APINFO, 0x02 );
		errlog_no = 0x1a ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}
	HFC_EXIT("read_adap_info") ;

	return( 0 );
}


/*
 * Function:    read_target_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read target_info) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *    
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 * Notes:       
 */
int read_target_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	struct target_info *target;		/* Retrieval target */

	uchar	errlog_no ;

	HFC_ENTRY("read_target_info") ;

	/*--------------------------------------------------------------*
	 * Specified check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_TGINFO, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct target_info) ) {	/* Is size short for target_info? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_TGINFO, 0x01 );
		return( EINVAL );
	}
	/*--------------------------------------------------------------*
	 * Target search
	 *--------------------------------------------------------------*/
	target = NULL;
	if ( diag->uni.dev_target.ww_name ) {
		/* WWN?*/
		target = hfc_hash_target_info_wwn( ap, diag->uni.dev_target.ww_name );
	}
	else {
		/* Scsi_id? */
		target = hfc_hash_target_info( ap, (uint)diag->uni.dev_target.target_id );
	}

	if ( target == NULL ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_TGINFO, 0x02 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)target, (char *)(ulong)diag->addr, sizeof(struct target_info) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_TGINFO, 0x03 );
		errlog_no = 0x1b ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_target_info");

	return( 0 );
}


/*
 * Function:    read_manage_info
 *
 * Purpose:     Processing for ioctl(HFCDIAG0) subcommand (Read manage_info) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 * Notes:       
 */
int read_manage_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_manage_info");

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MNG, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct manage_info) ) {	/* Is size short for manage_info? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MNG, 0x01 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data Copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (void *)&hfc_manage_info, (void *)(ulong)diag->addr, sizeof(struct manage_info) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MNG, 0x02 );
		errlog_no = 0x1d ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_manage_info") ;

	return( 0 );
}


/*
 * Function:    read_hfctrace
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read hfctrace)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     0		Normal end
 *             EINVAL	Specification error
 * Notes:       
 */
int read_hfctrace(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_hfctrace");

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCTRC, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct hfctrace) ) {	/* Is size short for hfctrace? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCTRC, 0x01 );
		return( EINVAL );
	}

	if( (diag->uni.xob_xrb_scmd.no >= ap->trc_max) ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCTRC, 0x02 );
		return( EINVAL );
	}

	if( diag->uni.xob_xrb_scmd.cnt != MAX_READ_CNT ) {	/* Read number */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCTRC, 0x03 );
		return( EINVAL );
	}

	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)(ap->trc_ptr + diag->uni.xob_xrb_scmd.no),
		(char *)(ulong)diag->addr, sizeof(struct hfctrace) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HFCTRC, 0x04 );
		errlog_no = 0x1e ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	diag->uni.xob_xrb_scmd.current_no = ap->trc_num ;

	HFC_EXIT("read_hfctrace");

	return( 0 );
}


/*
 * Function:    read_mailbox
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read mailbox) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 *
 * Notes:       
 */
int read_mailbox(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;

	HFC_ENTRY("read_mailbox") ;
	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MB, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(struct mailbox) ) {	/* Is size short for mailbox? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MB, 0x01 );
		return( EINVAL );
	}
	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)ap->mb, (char *)(ulong)diag->addr, sizeof(struct mailbox) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_MB, 0x02 );
		errlog_no = 0x1f ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_mailbox") ;

	return( 0 );
}


/*
 * Function:    read_version
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read version)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 * Notes:       
 */
int read_version(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	uchar	errlog_no ;


	HFC_ENTRY("read_version");

	/*--------------------------------------------------------------*
	 * Specification check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_VER, 0x00 );
		return( EINVAL );
	}
	if( diag->length < sizeof(hfc_ver) ) {	/* Is size short for hfc_ver? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_VER, 0x01 );
		return( EINVAL );
	}
	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)hfc_ver, (char *)(ulong)diag->addr, sizeof(hfc_ver) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_VER, 0x02 );
		errlog_no = 0x20 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	HFC_EXIT("read_version") ;

	return( 0 );
}


/*
 * Function:    read_hwlog
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read hw_log) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail
 * Notes:       
 */
int read_hwlog(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{

	uchar	errlog_no ;

	HFC_ENTRY("read_hwlog") ;

	/*--------------------------------------------------------------*
	 * Specified check
	 *--------------------------------------------------------------*/
	if( !diag->addr ) {	/* Is virtual address unassigned? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HWLOG, 0x00 );
		return( EINVAL );
	}
	if( diag->length < HFC_HWLOG_SIZE ) {	/* Is size short for hw_log? */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HWLOG, 0x01 );
		return( EINVAL );
	}
	/*--------------------------------------------------------------*
	 * Data copy
	 *--------------------------------------------------------------*/
	if ( COPYOUT( (char *)(ap->mp_adap_info->hw_log), (char *)(ulong)diag->addr, HFC_HWLOG_SIZE ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_HWLOG, 0x02 );
		errlog_no = 0x21 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
		return ( EFAULT );
	}

	set_bit(HFC_HWLOG_VALID , (ulong *)&ap->io_status);

	HFC_EXIT("read_hwlog") ;

	return( 0 );
}


/*
 * Function:    read_tree
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read tree)
 *               - Response data is device configuration information
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail 
 * Notes:       
 */
int read_tree(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	struct a_info	adap ;
	struct t_info	tg ;
	struct t_info	*t_ptr ;
	struct m_info	mp ;
	struct target_info *target ;

	uint			entry_num , entry_num_wk ;
	uint			i ;

	HFC_ENTRY("read_tree") ;

	diag->uni.dev_tree.entry_num = 0 ;

	if( diag->uni.dev_tree.type == HFC_DGRD_MPINFO ){
		if( diag->length < sizeof( struct m_info ) ){
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_TREE, 0x00 );
			return( EINVAL );
		}
		BZERO((char *)&mp, sizeof(struct m_info));

		mp.lock = ap->mp_adap_info->lock;										/* @0005 */
		mp.status = ap->mp_adap_info->status;									/* @0005 */
		mp.mck_result = ap->mp_adap_info->mck_result;							/* @0005 */
		mp.port_cnt = ap->mp_adap_info->port_cnt;								/* @0005 */
		mp.mck_seq_no = ap->mp_adap_info->mck_seq_no;							/* @0005 */
		mp.sys_rev = ap->mp_adap_info->sys_rev;									/* @0005 */

		if ( COPYOUT( (char *)&mp, (char *)(ulong)diag->addr, sizeof(struct m_info) ) != 0 )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_TREE, 0x02 );
			return ( EFAULT );
		}
		diag->uni.dev_tree.entry_num = 1 ;

		return( 0 );
	}

	if( diag->uni.dev_tree.type == HFC_DGRD_ADAP )
	{
		if( diag->length < sizeof(struct a_info) )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_TREE, 0x00 );
			return(EINVAL) ;
		}
		adap.status = ap->status ;
		adap.scsi_id = ap->scsi_id ;
		adap.ww_name = ap->ww_name ;
		adap.node_name = ap->node_name ;
		adap.connect_type = ap->connect_type ;
		adap.max_data_rate = ap->max_data_rate ;
		adap.port_no = ap->port_no ;
		adap.attach_status = ap->attach_status ;
		adap.open_status = ap->open_status ;
		adap.mb_status = ap->mb_status ;
		adap.mb_resp = ap->mb_resp ;
		adap.mb_results = ap->mb_results;
		adap.mb_retry_cnt = ap->mb_retry_cnt ;
		adap.pkt_no = ap->pkt_no ;
		adap.pkt_cnt = ap->pkt_cnt ;
		adap.xob_no = ap->xob_no ;
		adap.xrb_no = ap->xrb_no ;
		adap.iov_no = ap->iov_no ;
		adap.iov_map_cnt = ap->iov_map_cnt ;
		adap.target_cnt = ap->target_cnt ;
		adap.xob_exec_cnt = ap->xob_exec_cnt ;
		adap.xob_wait_exec_cnt = ap->xob_wait_exec_cnt ;
		for(i=0 ; i<MAX_FRAME_CNT ; i++)
		{
			adap.save_xob_outp[i] = ap->save_xob_outp[i] ;
			adap.xob_outp_end[i]  = ap->xob_outp_end[i] ;
		}
		adap.curr_xob_outp = ap->curr_xob_outp ;
		adap.frame_chkp = ap->frame_chkp ;
		adap.frame_inp = ap->frame_inp ;
		adap.scsi_exec_cnt = ap->scsi_exec_cnt ;
		adap.scsi_end_cnt = ap->scsi_end_cnt ;
		adap.scsi_err_cnt = ap->scsi_err_cnt ;
		adap.link_dead_cnt = ap->link_dead_cnt ;
		adap.pci_err_cnt = ap->pci_err_cnt ;
		adap.mck_err_cnt = ap->mp_adap_info->mck_err_cnt ;	/* FCLNX-GPL-057 */
		adap.io_status = ap->io_status ;
		adap.trc_num = ap->trc_num ;
		adap.next_dstart_cnt = ap->next_dstart_cnt ;

		if ( COPYOUT( (char *)&adap, (char *)(ulong)diag->addr, sizeof(struct a_info) ) != 0 )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_TREE, 0x00 );
			return ( EFAULT );
		}

		diag->uni.dev_tree.entry_num = 1 ;

		HFC_EXIT("read_tree");
		return( 0 );
	}

	if( diag->uni.dev_tree.type == HFC_DGRD_TARGET )
	{
		entry_num = diag->length / sizeof(struct t_info) ;
		if( entry_num == 0 )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_TREE, 0x00 );
			return(EINVAL) ;
		}
		entry_num_wk = 0 ;

		t_ptr = (struct t_info *)(ulong)diag->addr ;

		for(i=0; i<MAX_TARGET_PROBE; i++)
		{
			if( ap->target_arg[i] == NULL )
				continue ;
			target = ap->target_arg[i] ;

			while( (target != NULL) && (entry_num_wk < entry_num) )
			{
				if( !test_bit(HFC_TARGETINF_VALID , (ulong *)&target->flags) )
				{
					target = target->next ;
					continue ;
				}
				tg.status = target->status ;
				tg.flags = target->flags ;
				tg.device_flags = target->device_flags ;
				tg.pseq = target->pseq ;
				tg.target_id = target->target_id ;
				tg.scsi_id = target->scsi_id ;
				tg.ww_name = target->ww_name ;
				tg.node_name = target->node_name ;
				tg.we_que_cnt = target->we_que_cnt ;
				tg.wx_que_cnt = target->wx_que_cnt ;
				tg.next_dstart_flag = target->next_dstart_flag ;

				if ( COPYOUT( (char *)&tg, (char *)t_ptr, sizeof(struct t_info) ) != 0 )
				{
					HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
							HFC_TRC_IOCTL_TREE, 0x00 );
					diag->uni.dev_tree.entry_num = entry_num_wk ;

					return ( EFAULT );
				}
				entry_num_wk++ ;
				t_ptr++ ;
				target = target->next ;
			}
		}
		diag->uni.dev_tree.entry_num = entry_num_wk ;

		HFC_EXIT("read_tree");

		return( 0 );
	}
	HFC_EXIT("read_tree");

	return( 0 );
}


/*
 * Function:    read_dev_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read dev_info)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT   Copy process fail  *
 * Notes:       
 */
int read_dev_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int ret;
	
	if(hfc_manage_info.hfcldd_mp_mod) {
		ret = hfc_manage_info.npubp->read_dev_info_mp(ap, diag);
	}
	else {
		return (EIO);
	}
	return (ret);
}


/*
 * Function:    read_lg_target_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read lg_target_info)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  EIO        - Other errors
 * Notes:       
 */
int read_lg_target_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int ret;
	
	if(hfc_manage_info.hfcldd_mp_mod) {
		ret = hfc_manage_info.npubp->read_lg_target_info_mp(ap, diag);
	}
	else {
		return (EIO);
	}
	return (ret);
}


/*
 * Function:    read_lg_dev_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read lg_target_info) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  EIO        - Other errors
 * Notes:       
 */
int read_lg_dev_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int ret;
	
	if(hfc_manage_info.hfcldd_mp_mod) {
		ret = hfc_manage_info.npubp->read_lg_dev_info_mp(ap, diag);
	}
	else {
		return (EIO);
	}
	return (ret);
}


/*
 * Function:    read_lg_path_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read lg_target_info) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  EIO        - Other errors
 * Notes:       
 */
int read_lg_path_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int ret;
	
	if(hfc_manage_info.hfcldd_mp_mod) {
		ret = hfc_manage_info.npubp->read_lg_path_info_mp(ap, diag);
	}
	else {
		return (EIO);
	}
	return (ret);
}


/*
 * Function:    read_lg_path_info1
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read lg_path_info1) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  EIO        - Other errors
 * Notes:       
 */
int read_lg_path_info1(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int ret;
	
	if(hfc_manage_info.hfcldd_mp_mod) {
		ret = hfc_manage_info.npubp->read_lg_path_info1_mp(ap, diag);
	}
	else {
		return (EIO);
	}
	return (ret);
}


/*
 * Function:    read_lg_path_info2
 *
 * Purpose:     Process for ioctl(HFCDIAG0) ubcommand (Read lg_path_info2)
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  EIO        - Other errors
 * Notes:       
 */
int read_lg_path_info2(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int ret;
	
	if(hfc_manage_info.hfcldd_mp_mod) {
		ret = hfc_manage_info.npubp->read_lg_path_info2_mp(ap, diag);
	}
	else {
		return (EIO);
	}
	return (ret);
}


/*
 * Function:    read_failover_info
 *
 * Purpose:     Process for ioctl(HFCDIAG0) subcommand (Read failover_info) 
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:     
 *  0          - Normal end
 *  EIO        - Other errors
 * Notes:       
 */
int read_failover_info(
	struct	adap_info	*ap,	/* struct adap_info */
	struct	diag_ioctl	*diag)	/* struct diag_ioctl */
{
	int ret;
	
	if(hfc_manage_info.hfcldd_mp_mod) {
		ret = hfc_manage_info.npubp->read_failover_info_mp(ap, diag);
	}
	else {
		return (EIO);
	}
	return (ret);
}


/*
 * Function:    hfc_diag
 *
 * Purpose:     Process each ioctl(HFCDIAG0) subcommand
 *
 * Arguments:   
 *  ap         - struct adap_info
 *  diag       - struct diag_ioctl
 *
 * Returns:    0		Normal end
 *             EINVAL	Specification error
 *             EFAULT	Copy process fail
 *
 * Notes:      - Be sure to lock (spin_lock) adap_info->tbl_lock
 *             before calling this subroutine.
 */
int hfc_diag(
	void	*arg,				/* struct diag_ioctl */
	struct	adap_info	*ap)	/* struct adap_info */
{

	struct	diag_ioctl	diag;	/* DIAD internal area */
	struct  mp_adap_info *mpap; /* mp_adap_info area  */
	int		rtn ;
	uint	i = 0 ;
	uchar	errlog_no ;

	HFC_ENTRY("hfc_diag") ;
	/*--------------------------------------------------------------*
	* Copy internal DIAG arguments from caller
	*--------------------------------------------------------------*/
	if ( COPYIN( (char *)arg, (char *)&diag, sizeof(struct diag_ioctl) ) != 0 ) {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_DIAG, 0x00 );
		errlog_no = 0x03 ;
		hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5E, &errlog_no, 1) ;
		return ( EFAULT );
	}
	
#if defined(__x86_64) && !defined(_HFC_FCGW) &&  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
	/* Lower 32bit is effective for pointers */
	if(ap->ioctl32) {
		diag.addr = diag.addr & 0xffffffffU;
	}
#endif
	
	if( ap -> open_status != HFC_OPENED )
	{
		/* Designated adapter has not opened. */
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_DIAG, 0x01 ); 

		return( EIO );
	}

	mpap = ap->mp_adap_info;
	if( mpap == NULL ) {
		/* No corresponding mp_adap_info -> error */
		return( EIO );
	}

	if(!(test_bit(HFC_ATTACH, (ulong *)&ap->attach_status ) ) ) {				//FCLNX-0294
		if (!((diag.sub_cmd == HFC_DGRD_ADAP)
		   || (diag.sub_cmd == HFC_DGRD_TARGET)
		   || (diag.sub_cmd == HFC_DGRD_VER)
		   || (diag.sub_cmd == HFC_DGRD_TREE)
		   || (diag.sub_cmd == HFC_DG_RDPCI)
		   || (diag.sub_cmd == HFC_DG_WRPCI)
		   || (diag.sub_cmd == HFC_DG_RDPCI_CFG)
		   || (diag.sub_cmd == HFC_DG_WRPCI_CFG)
		   || (diag.sub_cmd == HFC_DGRD_MPINFO)
		   || (diag.sub_cmd == HFC_DGRD_MNGINFO))) {
			/* Is adapter disable? */
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_DIAG, 0x02 );
			return( EIO );
		}
	}

	/*----------------------------------------------------------------
	 * Read F/W operation mode and check
	 *---------------------------------------------------------------*/

	ap->fcp_mode = hfc_read_reg(ap, HFC_IOSPACE_CA_FLAG, PCI_LENGTH_01);

	i = 0 ;
	while( diag_check[i].sub_command != 0 )
	{
		if( diag_check[i].sub_command == diag.sub_cmd )
			break ;
		i++ ;
	}
	if ( ap->pkg.type == HFC_PKTYPE_FPP ) {
		if( !(diag_check[i].support_pk & HFC_DIAG_FPP) )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_DIAG, 0x03 );  
			return( EINVAL );
		}
	}
	else if( ap->pkg.type == HFC_PKTYPE_FIVE )
	{
		if( !(diag_check[i].support_pk & HFC_DIAG_FIVE) )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_DIAG, 0x04 ); 
			return( EINVAL );
		}
	}
	else {
		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_DIAG, 0x05 ); 
		return( EINVAL );
	}
	if( ap->fcp_mode == FW_MODE_NORMAL )
	{
		if ( ap->pkg.type == HFC_PKTYPE_FPP ) {
			if( !(diag_check[i].flags_fpp & HFC_DIAG_NORMAL_MODE) )
			{
				HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
						HFC_TRC_IOCTL_DIAG, 0x06 ); 
				return( EINVAL );
			}
		} 
	}
	if( ap->fcp_mode == FW_MODE_DIAG )
	{
		if( !(diag_check[i].flags_fpp & HFC_DIAG_DIAG_MODE) )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_DIAG, 0x07 ); 
			return( EINVAL );
		}
	}
	if( (ap->fcp_mode == FW_MODE_NAI_LOOP) || (ap->fcp_mode == FW_MODE_GAI_LOOP) )
	{
		if( !(diag_check[i].flags_fpp & HFC_DIAG_LOOP_MODE) )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_DIAG, 0x08 );
			return( EINVAL );
		}
	}
	if( ap->fcp_mode == FW_MODE_IOS )
	{
		if( !(diag_check[i].flags_fpp & HFC_DIAG_IOS_MODE) )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_DIAG, 0x09 ); 
			return( EINVAL );
		}
	}

	switch( diag.sub_cmd )
	{
		case HFC_DG_SETFW :
		case HFC_DG_MIHLOG :
		case HFC_DG_LDCHTRC :
		case HFC_DG_FORCSLOG :
		case HFC_DG_FWSTART :
		case HFC_DG_FWPOST :
		case HFC_DG_RDPCI :
		case HFC_DG_WRPCI :
		case HFC_DG_RDPCI_CFG :
		case HFC_DG_WRPCI_CFG :
		case HFC_DG_FSTOP :
		case HFC_DG_INITMDST :
		case HFC_DG_FCPMDST :
			HFC_ADAP_LOCK(mpap,HFC_MP_ADAP_BUSY);
			if( test_bit(HFC_IOCTL_WR_CHECK, (ulong *)&mpap->status) )
			{
				HFC_ADAP_UNLOCK(mpap,HFC_MP_ADAP_BUSY);
				return -EBUSY ;
			}
			else{
				set_bit(HFC_IOCTL_WR_CHECK, (ulong *)&mpap->status);
			}
			HFC_ADAP_UNLOCK(mpap,HFC_MP_ADAP_BUSY);
			break;
		default:
			break;

	}

	rtn = diag_check[i].cmd_func( ap , &diag) ;

	switch( diag.sub_cmd )
	{
		case HFC_DG_SETFW :
		case HFC_DG_MIHLOG :
		case HFC_DG_LDCHTRC :
		case HFC_DG_FORCSLOG :
		case HFC_DG_FWSTART :
		case HFC_DG_FWPOST :
		case HFC_DG_RDPCI :
		case HFC_DG_WRPCI :
		case HFC_DG_RDPCI_CFG :
		case HFC_DG_WRPCI_CFG :
		case HFC_DG_FSTOP :
		case HFC_DG_INITMDST :
		case HFC_DG_FCPMDST :
			if( mpap != NULL ){
				if( test_bit(HFC_IOCTL_WR_CHECK, (ulong *)&mpap->status) )
				{
					HFC_ADAP_LOCK(mpap,HFC_MP_ADAP_BUSY);
					clear_bit(HFC_IOCTL_WR_CHECK, (ulong *)&mpap->status);
					HFC_ADAP_UNLOCK(mpap,HFC_MP_ADAP_BUSY);
				}
			}
			break;
		default:
			break;
	}

	if( rtn != 0 ) {	/* Does any of this routine have an error? */

		HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
				HFC_TRC_IOCTL_DIAG, 0x0a ); 
		return( rtn );
	}

	/*--------------------------------------------------------------*
	 * Copy back internal diag arguments to caller 
	 *  (only for diag_ioctl resp)
	 *--------------------------------------------------------------*/
	if( (diag_check[i].flags_fpp & HFC_DIAG_ARG_RESP) ||
		(diag_check[i].flags_five & HFC_DIAG_ARG_RESP) )
	{
		if ( COPYOUT( (char *)&diag, (char *)arg, sizeof(struct diag_ioctl) ) != 0 )
		{
			HFC_DBGPRT("ioctl error(trcid=0x%04x, subid=0x%04x) \n",
					HFC_TRC_IOCTL_DIAG, 0x0b ); 
			errlog_no = 0x0 ;
			hfc_errlog(ap, NULL, NULL, HFC_ERRLOG_TYPE_NONE, ERRID_HFCP_ERRC, 0x5F, &errlog_no, 1) ;
			return ( EFAULT );
		}
	}

	HFC_EXIT("hfc_diag") ;

	return( 0 );	/* Normal end */
}
